@@ -17,13 +17,14 @@ type Version struct {
 var (
 	// DelveVersion is the current version of Delve.
 	DelveVersion = Version{
-		Major: "1", Minor: "7", Patch: "3", Metadata: "",
+		Major: "1", Minor: "8", Patch: "0", Metadata: "",
 		//TODO(aarzilli): before updating this to 1.8.0 re-enable staticcheck test
-		Build: "$Id: 4c81a2a322716591167f0bd690f7ef656781b6f4 $",
+		Build: "$Id: 6a6c9c332d5354ddf1f8a2da3cc477bd18d2be53 $",
 	}
 )
 
 func (v Version) String() string {
+	fixBuild(&v)
 	ver := fmt.Sprintf("Version: %s.%s.%s", v.Major, v.Minor, v.Patch)
 	if v.Metadata != "" {
 		ver += "-" + v.Metadata
@@ -38,3 +39,7 @@ var buildInfo = func() string {
 func BuildInfo() string {
 	return fmt.Sprintf("%s\n%s", runtime.Version(), buildInfo())
 }
+
+var fixBuild = func(v *Version) {
+	// does nothing
+}
